<?php
namespace CodingMs\AmazonPayment\Controller;

/***************************************************************
 *
 *  Copyright notice
 *
 *  (c) 2017 Thomas Deuling <typo3@coding.ms>, coding.ms
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use CodingMs\AmazonPayment\Domain\Model\PaymentPreset;
use CodingMs\AmazonPayment\Utility\Tools;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Extbase\Configuration\ConfigurationManagerInterface;
use TYPO3\CMS\Extbase\Utility\DebuggerUtility;
use TYPO3\CMS\Extbase\Utility\LocalizationUtility;

/**
 * AmazonController
 */
class AmazonController extends \TYPO3\CMS\Extbase\Mvc\Controller\ActionController
{

    /**
     * @var \CodingMs\AmazonPayment\Domain\Repository\PaymentPresetRepository
     * @inject
     */
    protected $paymentPresetRepository;

    /**
     * @var \CodingMs\AmazonPayment\Service\PaymentService
     * @inject
     */
    protected $paymentService;

    /**
     * JavaScript configuration
     * @var array
     */
    protected $javaScriptConfig = array();

    /**
     * @var bool
     */
    protected $debug = false;

    /**
     * Initialize every action
     */
    public function initializeAction()
    {
        // Debugging
        $this->debug = (bool)$this->settings['debug'];
        // JavaScript configuration
        $this->javaScriptConfig['debug'] = $this->debug;
        $this->javaScriptConfig['clientId'] = $this->settings['amazon']['clientId'];
        $this->javaScriptConfig['sellerId'] = $this->settings['amazon']['sellerId'];
        $this->javaScriptConfig['login']['type'] = $this->settings['amazon']['login']['type'];
        $this->javaScriptConfig['login']['color'] = $this->settings['amazon']['login']['color'];
        $this->javaScriptConfig['login']['size'] = $this->settings['amazon']['login']['size'];
        $this->javaScriptConfig['login']['language'] = $this->settings['amazon']['login']['language'];
        $this->javaScriptConfig['login']['scope'] = $this->settings['amazon']['login']['scope'];
        $this->javaScriptConfig['login']['popup'] = (bool)$this->settings['amazon']['login']['popup'];
        $this->javaScriptConfig['login']['redirectUrl'] = trim($this->settings['amazon']['login']['redirectUrl']);
        // Initialize parent
        parent::initializeAction();
    }

    /**
     * Inserts the checkout process
     */
    public function checkoutAction()
    {
        // Get payment preset
        $paymentPreset = 0;
        if ($this->request->hasArgument('paymentPreset')) {
            $paymentPreset = (int)$this->request->getArgument('paymentPreset');
        }
        /** @var $paymentPreset \CodingMs\AmazonPayment\Domain\Model\PaymentPreset */
        $paymentPreset = $this->paymentPresetRepository->findByIdentifier($paymentPreset);
        if ($paymentPreset instanceof PaymentPreset) {
            $this->view->assign('paymentPreset', $paymentPreset);
            /** @var \TYPO3\CMS\Extbase\Mvc\Web\Routing\UriBuilder */
            $uriBuilder = $this->controllerContext->getUriBuilder();
            $apiUri = $uriBuilder->reset()->setCreateAbsoluteUri(true)
                ->setUseCacheHash(false)
                ->setTargetPageType(1479812019)
                ->uriFor('api', array(), 'Amazon', 'AmazonPayment', 'JsonApi');
            // JavaScript configuration
            $this->javaScriptConfig['apiUri'] = $apiUri;
            $this->javaScriptConfig['paymentPresetUid'] = (int)$paymentPreset->getUid();
            $this->javaScriptConfig['addressRequired'] = $paymentPreset->isAddressRequired();
        }
        $this->view->assign('config', json_encode($this->javaScriptConfig));
    }

    /**
     * Displays selected payment presets
     */
    public function showPresetsAction()
    {
        $paymentPresets = array();
        $paymentPresetUids = GeneralUtility::trimExplode(',', $this->settings['paymentPresets'], true);
        if (!empty($paymentPresetUids)) {
            foreach ($paymentPresetUids as $paymentPresetUid) {
                $paymentPreset = $this->paymentPresetRepository->findByIdentifier($paymentPresetUid);
                if ($paymentPreset instanceof PaymentPreset) {
                    $paymentPresets[] = $paymentPreset;
                }
            }
        } else {
            $paymentPresets = $this->paymentPresetRepository->findAll();
        }
        $this->settings['wrapperCssClass'] = GeneralUtility::camelCaseToLowerCaseUnderscored($this->settings['templateSection']);
        $this->settings['wrapperCssClass'] = str_replace('_', '-', $this->settings['wrapperCssClass']);
        $this->view->assign('settings', $this->settings);
        $this->view->assign('paymentPresets', $paymentPresets);
    }

    /**
     * @return string
     */
    public function apiAction()
    {
        $jsonArray = array();
        $region = '';
        $validRegions = array('us', 'de', 'uk', 'jp');
        /**
         * @todo validate; must be passed by argument
         */
        $region = 'de';
        // Get the payment action which must be performed
        $paymentAction = '';
        $validPaymentActions = array('initialize', 'confirm', 'finish', 'cancel');
        if ($this->request->hasArgument('paymentAction')) {
            $paymentAction = trim($this->request->getArgument('paymentAction'));
        }
        if (!in_array($paymentAction, $validPaymentActions)) {
            $jsonArray['error']['message'] = 'Invalid payment action';
        }
        // Get address access token
        $addressConsentToken = '';
        if ($this->request->hasArgument('addressConsentToken')) {
            $addressConsentToken = $this->request->getArgument('addressConsentToken');
        }
        // Always get order reference id
        // and initialize the payment service
        $orderReferenceId = '';
        if ($this->request->hasArgument('orderReferenceId')) {
            // Get order reference id
            $orderReferenceId = $this->request->getArgument('orderReferenceId');
            // Initialize the payment service
            $config = array(
                'merchant_id' => $this->settings['amazon']['sellerId'],
                'access_key' => $this->settings['amazon']['accessKey'],
                'secret_key' => $this->settings['amazon']['secretKey'],
                'client_id' => $this->settings['amazon']['clientId'],
                'region' => $region,
                'sandbox' => (bool)$this->settings['amazon']['sandbox']
            );
            // Proxy
            $proxy = array();
            if((bool)$this->settings['amazon']['proxy']) {
                $proxy = array(
                    'proxy_user_host' => $this->settings['amazon']['proxyHost'],
                    'proxy_user_port' => $this->settings['amazon']['proxyPort'],
                    'proxy_user_name' => $this->settings['amazon']['proxyUsername'],
                    'proxy_user_password' => $this->settings['amazon']['proxyPassword'],
                );
            }
            // Initialize payment service
            $newRecordPersisted = $this->paymentService->initialize($config, $orderReferenceId, $proxy);
            // Get address consent token
            // in order to save user information in payment
            // But only, when payment record was just persisted
            if ($addressConsentToken != '' && $newRecordPersisted) {
                // Save user information in payment object
                $response = $this->paymentService->getUserInfo($addressConsentToken);
                if (isset($response['Error'])) {
                    $jsonArray['error'] = $response['Error'];
                }
            }
        }
        // Initialize the payment itself
        if ($paymentAction == 'initialize' && !isset($jsonArray['error'])) {
            // Get payment preset - absolutely required
            // The payment preset contains all the necessary payment information
            $paymentPreset = null;
            if ($this->request->hasArgument('paymentPresetUid')) {
                $paymentPresetUid = (int)$this->request->getArgument('paymentPresetUid');
                /** @var $paymentPreset \CodingMs\AmazonPayment\Domain\Model\PaymentPreset */
                $paymentPreset = $this->paymentPresetRepository->findByIdentifier($paymentPresetUid);
                if (!($paymentPreset instanceof PaymentPreset)) {
                    $paymentPreset = null;
                    $jsonArray['error'] = 'Invalid payment preset';
                } else {
                    // Set order reference details
                    $responseArray = $this->paymentService->setOrderReferenceDetails(array(
                        'merchant_id' => $this->settings['amazon']['sellerId'],
                        'amazon_order_reference_id' => $orderReferenceId,
                        'amount' => $paymentPreset->getAmount(),
                        'currency_code' => $paymentPreset->getCurrencyCode(),
                        'platform_id' => 'AI1XRQMUB41F',
                        'seller_note' => $paymentPreset->getSellerNote(),
                        'seller_order_id' => $this->paymentService->getPaymentOrderId($this->settings['amazon']['orderIdPrefix']),
                        'store_name' => $this->settings['amazon']['storeName']
                    ), $paymentPreset);
                    if (isset($responseArray['Error'])) {
                        $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
                    }
                }
            }
            // First of all, get the current order reference details
            $responseArray = $this->paymentService->getOrderReferenceDetails(array(
                'amazon_order_reference_id' => $orderReferenceId,
                'address_consent_token' => $addressConsentToken
            ));
            if (isset($responseArray['Error'])) {
                $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
            } else {
                // Extract order details
                $orderReferenceDetails = $responseArray['GetOrderReferenceDetailsResult']['OrderReferenceDetails'];
                $jsonArray['orderReferenceDetails'] = $this->paymentService->extractOrderReferenceDetails($orderReferenceDetails, $this->debug);
            }
        } else if ($paymentAction == 'confirm' && !isset($jsonArray['error'])) {
            // Confirm payment data
            // First of all, get the current order reference details
            $responseArray = $this->paymentService->getOrderReferenceDetails(array(
                'amazon_order_reference_id' => $orderReferenceId,
                'address_consent_token' => $addressConsentToken
            ));
            if (isset($responseArray['Error'])) {
                $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
            } else {
                // Extract order details
                $orderReferenceDetails = $responseArray['GetOrderReferenceDetailsResult']['OrderReferenceDetails'];
                $jsonArray['orderReferenceDetails'] = $this->paymentService->extractOrderReferenceDetails($orderReferenceDetails, $this->debug);
                // Check for allowed shipping countries
                $countryCode = $jsonArray['orderReferenceDetails']['shippingAddress']['countryCode'];
                $allowedShippingCountries = $this->settings['allowedShippingCountries'];
                // Check if there are some constraints
                if (isset($jsonArray['orderReferenceDetails']['constraints'])) {
                    $jsonArray['html'] = '';
                } else if (!array_key_exists($countryCode, $allowedShippingCountries) && $this->paymentService->isAddressRequired()) {
                    $translationKey = 'tx_amazonpayment_message.invalid_shipping_country';
                    $jsonArray['html'] = '';
                    $jsonArray['orderReferenceDetails']['constraints'] = array(
                        'InvalidShippingCounty' => LocalizationUtility::translate($translationKey, 'AmazonPayment')
                    );
                    /**
                     * @todo: We need to set the shipping cost at any place!
                     */
                } else {
                    // No constraints..
                    // Render confirm information
                    $jsonArray['html'] = $this->renderConfirmTemplate('Amazon/Confirm.html',
                        $jsonArray['orderReferenceDetails']);
                }
            }
        } else if ($paymentAction == 'finish' && !isset($jsonArray['error'])) {
            // Finish payment
            // First of all, get the current order reference details
            $responseArray = $this->paymentService->getOrderReferenceDetails(array(
                'amazon_order_reference_id' => $orderReferenceId,
                'address_consent_token' => $addressConsentToken
            ));
            if (!isset($responseArray['Error'])) {
                // Extract order details
                $orderReferenceDetails = $responseArray['GetOrderReferenceDetailsResult']['OrderReferenceDetails'];
                $jsonArray['orderReferenceDetails'] = $this->paymentService->extractOrderReferenceDetails($orderReferenceDetails, $this->debug);
                // Order state must be in draft or were suspended!
                $orderState = $jsonArray['orderReferenceDetails']['orderState'];
                // Check if there are no constraints left
                $constraints = isset($jsonArray['orderReferenceDetails']['constraints']) ? $jsonArray['orderReferenceDetails']['constraints'] : false;
                // After a declined payment, there is a:
                // The buyer has not been able to select a Payment method for the given Order Reference.
                // Code: PaymentPlanNotSet
                if(is_array($constraints)) {
                   // Constraints will be displayed in frontend!
                }
                // No constraints left and state is draft or suspended
                else if (($orderState == 'Draft' || $orderState == 'Suspended')  && !$constraints) {
                    //
                    // Try to confirm the order reference
                    $responseArray = $this->paymentService->confirmOrderReference(array(
                        'merchant_id' => $this->settings['amazon']['sellerId'],
                        'amazon_order_reference_id' => $orderReferenceId,
                        'address_consent_token' => $addressConsentToken
                    ));
                    if (!isset($responseArray['Error'])) {
                        //
                        // Try to authorize the payment
                        $responseArray = $this->paymentService->authorize(array(
                            'merchant_id' => $this->settings['amazon']['sellerId'],
                            'amazon_order_reference_id' => $orderReferenceId,
                            'authorization_amount' => $this->paymentService->getAmount(),
                            'currency_code' => $this->paymentService->getCurrencyCode(),
                            'authorization_reference_id' => $this->paymentService->getPaymentOrderId($this->settings['amazon']['orderIdPrefix']),
                            'capture_now' => !$this->paymentService->isChargeOnShipment(),
                            'transaction_timeout' => 0
                        ));
                        if (!isset($responseArray['Error'])) {
                            // Extract authorization details
                            $authorizationDetails = $responseArray['AuthorizeResult']['AuthorizationDetails'];
                            $authorizationDetailsArray = $this->paymentService->extractAuthorizationDetails($authorizationDetails);
                            // Fetch some authorization details

                            $jsonArray['authorizationDetails'] = array(
                                'authorizationStatus' => $authorizationDetailsArray['authorizationStatus']
                            );
                            // Authorization successful?
                            // Captured amount is enough?
                            if ($authorizationDetailsArray['authorizationStatus']['state'] == 'Closed' &&
                                $authorizationDetailsArray['authorizationStatus']['reasonCode'] == 'MaxCapturesProcessed' &&
                                $authorizationDetailsArray['authorizationAmount']['currencyCode'] == $this->paymentService->getCurrencyCode() &&
                                $authorizationDetailsArray['authorizationAmount']['amount'] == $this->paymentService->getAmount()
                            ) {
                                //
                                // Close order reference
                                $responseArray = $this->paymentService->closeOrderReference(array(
                                    'merchant_id' => $this->settings['amazon']['sellerId'],
                                    'amazon_order_reference_id' => $orderReferenceId,
                                    'closure_reason' => 'Payment finished'
                                ));
                                if (isset($responseArray['Error'])) {
                                    $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
                                } else {
                                    $jsonArray['message'] = 'Payment successful';
                                    // Emit signal
                                    $this->emitAfterAmazonPaymentSuccessfulSignal(array(
                                        'orderReferenceId' => $orderReferenceId,
                                        'chargeOnShipment' => $this->paymentService->isChargeOnShipment()
                                    ));
                                }
                                //
                            }
                            else if($authorizationDetailsArray['authorizationStatus']['state'] == 'Open' && $this->paymentService->isChargeOnShipment()) {
                                // Is charge-on-shipment the authorization-status is returned as Open
                                // this means, everything is fine
                                // The Shop-Owner needs to catch the amount of charge in the SellerCentral
                                $jsonArray['message'] = 'Payment successful';
                                // Emit signal
                                $this->emitAfterAmazonPaymentSuccessfulSignal(array(
                                    'orderReferenceId' => $orderReferenceId,
                                    'chargeOnShipment' => $this->paymentService->isChargeOnShipment()
                                ));
                            }
                            else if($authorizationDetailsArray['authorizationStatus']['state'] == 'Declined') {
                                // Authorization was declined!
                                //
                                // ReasonCode can be:
                                // InvalidPaymentMethod
                                // AmazonRejected':
                                //     Amazon rejected the payment!
                                //     AmazonRejected indicates that the Authorize operation call was
                                //     declined because of a determination made by Amazon. The order is
                                //     moved to a Closed state, and you should decline the transaction.
                                //     New order reference state: Close
                                $reasonCode = $authorizationDetailsArray['authorizationStatus']['reasonCode'];
                                $jsonArray['error'] = array(
                                    'Message' => Tools::translateErrorCode($reasonCode, $this->debug),
                                    'Code' => $reasonCode
                                );
                            }
                        }
                        //
                    } else {
                        // Transfer error messages
                        $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
                    }
                    //
                } else {
                    // Transfer error messages
                    $jsonArray['error'] = array(
                        'Message' => 'Invalid order state!',
                        'Code' => ''
                    );
                }
            } else {
                // Transfer error messages
                $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
            }
        } else if ($paymentAction == 'cancel' && !isset($jsonArray['error'])) {
            // Cancel payment
            $orderReferenceStatus = $this->paymentService->getAmazonOrderReferenceStatus();
            // But only on API side, if state is not Draft
            if($orderReferenceStatus != 'Draft') {
                $responseArray = $this->paymentService->cancelOrderReference(array(
                    'amazon_order_reference_id' => $orderReferenceId,
                    'merchant_id' => $this->settings['amazon']['sellerId']
                ));
                if (isset($responseArray['Error'])) {
                    $jsonArray['error'] = Tools::processError($responseArray['Error'], $this->debug);
                } else {
                    $translationKey = 'tx_amazonpayment_message.payment_successfully_canceled';
                    $jsonArray['message'] = LocalizationUtility::translate($translationKey, 'AmazonPayment');
                }
            }
            else {
                // No canceling required
                // Just logout on client side
                $translationKey = 'tx_amazonpayment_message.payment_successfully_canceled';
                $jsonArray['message'] = LocalizationUtility::translate($translationKey, 'AmazonPayment');
            }
        } else if(!isset($jsonArray['error'])) {
            $jsonArray['error'] = 'Unknown action or invalid state';
            if($this->debug) {
                $jsonArray['error'] .= ' (\'' . $paymentAction . '\')';
            }
        }
        return json_encode($jsonArray);
    }

    /**
     * Emits a signal after a payment was successful
     *
     * @param array $signalParameter
     * @return void
     */
    protected function emitAfterAmazonPaymentSuccessfulSignal($signalParameter)
    {
        $signalSlotDispatcher = GeneralUtility::makeInstance('TYPO3\\CMS\\Extbase\\SignalSlot\\Dispatcher');
        $signalSlotDispatcher->dispatch(__CLASS__, 'afterAmazonPaymentSuccessful', array($signalParameter, $this));
		//Mail
		$mailsubject=$this->settings['amazon']['mailSubject'];
		$mailreceiver=$this->settings['amazon']['mailReceiver'];
		$mailsender=$this->settings['amazon']['mailSender'];
		$mailtext=$this->settings['amazon']['mailText'];
		
		$mail = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance('TYPO3\\CMS\\Core\\Mail\\MailMessage');
		$mail->setSubject($mailsubject);
		$mail->setFrom(array($mailsender));
		$mail->setTo(array($mailreceiver));
		$mail->addPart($mailtext, 'text/html', 'utf-8');
		$mail->send();
    }

    /**
     * Render the confirm template
     * @param $partialFileName
     * @param $orderReferenceDetails
     * @return string Rendered HTML
     */
    protected function renderConfirmTemplate($partialFileName, $orderReferenceDetails)
    {
        // Render confirm information
        $configurationTypeFramework = ConfigurationManagerInterface::CONFIGURATION_TYPE_FRAMEWORK;
        $configuration = $this->configurationManager->getConfiguration($configurationTypeFramework, 'AmazonPayment');
        $partialRootPath = Tools::getTemplatePath($configuration['view'], 'template', $partialFileName);
        /** @var \TYPO3\CMS\Fluid\View\StandaloneView $listItemView */
        $confirmView = GeneralUtility::makeInstance('TYPO3\\CMS\\Fluid\\View\\StandaloneView');
        $confirmView->setControllerContext($this->controllerContext);
        $confirmView->setTemplatePathAndFilename($partialRootPath . $partialFileName);
        $confirmView->setPartialRootPath($partialRootPath);
        $confirmView->assign('settings', $this->settings);
        $confirmView->assign('orderReferenceDetails', $orderReferenceDetails);
        return base64_encode($confirmView->render());
    }

}